#include<stdlib.h>
#include<stdio.h>
#include<unistd.h>
#include<errno.h>
#include<string.h>
#include<time.h>
#include<fcntl.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<ctype.h>
#include<sys/time.h>
#include<sys/wait.h>
#define TIMELEN 15
#define CALLLEN 80
#define RETVALLEN 10
#define LINELEN 200
#define MAXFILES 50

int fd_arr[MAXFILES] = {-1}; //It seems bad to hard code this, but LANL is 32 processes 
//and running 50 traces against a single device seems unlikely.
char *paths[MAXFILES];
unsigned long long curr_offset[MAXFILES] = {0};
unsigned long long filesize[MAXFILES] = {0};
char *prefix;
//What stats should we collect?
double read_IO = 0.0; // bytes read
double read_elapsed =0.0; //elapsed time of all reads
double write_IO = 0.0; //bytes written
//We can only count this from explicit write/writev calls and 
//not various syncs (though I suppose there's none of them in the traces.)
double write_elapsed = 0.0; //elapsed time of all writes
//Should we take individual timings? We'll need a linked list for that, 
//keeping elapsed time and bytes xfer'd, unless we specify a second storage device to 
//write out timing information.
struct timeval global_start[1], global_stop[1], read_start[1], read_stop[1], write_start[1], write_stop[1];
double global_elapsed;
int num_read = 0;
int num_write = 0;

void print_error(char *msg){
        printf("Error: %s - %s\n", msg, strerror(errno));
        //exit(-1);
}

double diff_elapsed(struct timeval start[1], struct timeval stop[1]){
        if( start->tv_usec > stop->tv_usec){
                return (double)(stop->tv_sec - start->tv_sec - 1)+((double)(1000000 + stop->tv_usec - start->tv_usec)/1000000.0);
        }
        else{
                return (double)(stop->tv_sec - start->tv_sec)+((double)(stop->tv_usec - start->tv_usec)/1000000.0);
        }
}

//Count the number of arguments in the call.
int parm_count(char call[]){
        int count = 1;
        char *mark = strpbrk(call, ",");
        while (mark != NULL){
                count++;
                mark = strpbrk(mark+1, ",");
        }
        return count;
}

int gather_fsstat(char call[], char retval[]){
        char callcopy[strlen(call)];
        strcpy(callcopy, call);
        char *callname=strtok(call, "(");
        #ifdef DEBUG
        printf("call name: %s\n", callname);
        printf("ret val: %s\n", retval);
        #endif
        if((strcmp(callname, "open")==0) && (atoi(retval) > 0)){
                #ifdef DEBUG
                printf("open case\n");
                #endif
                int old_fd = atoi(retval);
                if(old_fd >=50){ print_error("Tried to open more than MAXFILES files. Increase the limit.\n"); }
                char *path_str = strtok(NULL, ","); //This gets parsed in funny, hence the next couple of lines.
                int pathlen = strlen(path_str)-1; //Shave off the quotes but need to add a terminator.
                char path[pathlen];
                memcpy(path, path_str+1, pathlen-1);
                memcpy(&path[pathlen-1], "\0", 1);
                paths[old_fd] = (char *)malloc(pathlen);
                memcpy(paths[old_fd], path, pathlen);
                #ifdef DEBUG

                #endif
        }
        else if( (strcmp(callname, "close")==0) && (atoi(retval)==0) ){ 
		//We do file creation when a file is closed in the trace.
                int fd = atoi(strtok(NULL, ")"));
                if(paths[fd] != NULL){ //The LANL traces call close multiple, consecutive times on an fd successfully...
                //Create file at paths[fd].
                #ifdef DEBUG
                //printf("Path to create: %s\n", paths[fd]);
                #endif
                char path_copy[strlen(paths[fd])];
                strcpy(path_copy, paths[fd]);
                if(chdir(prefix) != 0 ) {print_error("Failed to change to sandbox\n"); }
                char *first = strtok(paths[fd], "/");
                char initial[strlen(prefix)+strlen(first)];
                strcpy(initial, prefix);
                strcat(initial, first);
                char *path_curr = initial;
                char *path_next = strtok(NULL, "/");
                #ifdef CREATION_DEBUG
                printf("Path to create: %s\n", path_copy);
                #endif
                while(path_next != NULL){ //While next is still valid, path_curr holds a directory name.
                        #ifdef CREATION_DEBUG
                        printf("curr: %s, next: %s\n", path_curr, path_next);
                        #endif
                        //Make a dir called path_curr
                        if( (mkdir(path_curr, S_IRWXU|S_IRWXO|S_IRWXG) != 0 ) && (errno != EEXIST) ){ print_error("mkdir failed\n"); }
                        #ifdef CREATION_DEBUG
                        printf("Made dir\n");
                        #endif
                        if(chdir(path_curr) != 0 ) {print_error("Failed to change to new directory\n"); }
                        path_curr = path_next;
                        path_next = strtok(NULL, "/");
                } //On exit of the loop, path_curr holds the last valid path part, so make a file of that name.
                FILE *new_file = fopen(path_curr, "r"); //We want to check if the file already exists first.
                if(new_file == NULL){ //Does not exist so make it.
                        #ifdef CREATION_DEBUG
                        printf("File %s does not exist\n", path_curr); sleep(3);
                        printf("Filesize will be %llu\n", filesize[fd]);
                        #endif
                        new_file = fopen(path_curr, "w");
                        int i;
                        for(i=0; i< filesize[fd]; i++){ fwrite("0", 1, 1, new_file); }
                        fclose(new_file);

                        #ifdef CREATION_DEBUG
                        printf("Created %s, with filesize %llu\n", path_copy, filesize[fd]);
                        #endif
                }
                else{ //The file already exists so compare current filesize with newly computed filesize
                        struct stat statbuf;
                        #ifdef CREATION_DEBUG
                        printf("File %s already exists, checking filesize\n", path_curr);
                        #endif
                        if(fstat(fileno(new_file), &statbuf) == 0){
                                fclose(new_file);
                                #ifdef CREATION_DEBUG
                                printf("curr size: %llu, computed: %llu\n", (unsigned long long)statbuf.st_size, filesize[fd]);
                                #endif
                                if((unsigned long long)statbuf.st_size < filesize[fd]){ //computed size is greater than current size so append the file

                                        unsigned long long diff = filesize[fd] - (unsigned long long)statbuf.st_size;
                                        #ifdef CREATION_DEBUG
                                        printf("appending %llu bytes\n", diff);
                                        #endif
                                        new_file = fopen(path_curr, "a");
                                        int i;
                                        for(i=0; i< diff; i++){ fwrite("0", 1, 1, new_file); }
                                        fclose(new_file);
                                }
                                #ifdef CREATION_DEBUG
                                else{ printf("%s is already large enough\n", path_curr); }
                                #endif
                        }
                        else{ print_error("fstat failed\n"); }

                }
		//And clear the stats associated with that fd.
                free(paths[fd]);
                paths[fd] = NULL;
                curr_offset[fd] = 0;
                filesize[fd] = 0;
                if(chdir(prefix) != 0 ) {print_error("Failed to change to target\n"); }
                }
        }
        else if(( (strcmp(callname, "read")==0) || (strcmp(callname, "readv")==0) || (strcmp(callname, "write")==0) || (strcmp(callname, "writev")==0))
                 && (atoi(retval) >= 0) ){
                #ifdef DEBUG
                printf("read/readv/write/writev case\n");
                #endif
                int fd = atoi(strtok(NULL, ","));
		if(paths[fd] != NULL){
			#ifdef DEBUG
			printf("fd: %d\n", fd);
			printf("io to %s\n", paths[fd]);
			#endif
	                curr_offset[fd] += atoi(retval);
        	        if(curr_offset[fd] > filesize[fd]){
                	        filesize[fd] = curr_offset[fd];
                	        #ifdef DEBUG
                	        printf("%s filesize updated to %llu\n", paths[fd], filesize[fd]);
                       		 #endif
                	}
		}
		#ifdef DEBUG
		else{ printf("io to a std\n");} //Sometimes, this happens, I/O to a stdin/out/err.
		#endif
        }
        else if( (strcmp(callname, "_llseek") == 0) && (atoi(retval) >= 0) ){ 
		//Seeks normally return the resulting offset but return zeros in the traces.
                #ifdef DEBUG
                printf("_llseek case\n");
                #endif
                int fd = atoi(strtok(NULL, ","));
                char *offset_hi_str = strtok(NULL, ",");
                char *offset_lo_str = strtok(NULL, ",");
                strtok(NULL, ","); //Was result_str
                int whence = atoi(strtok(NULL, ")"));
                unsigned long offset_hi = strtoul(offset_hi_str, NULL, 10); 
		//I'm guessing this argument is a decimal.
                unsigned long offset_lo  = strtoul(offset_lo_str, NULL, 16); 
		//But it's strange that this one is given in hex.
                //unsigned long result = strtoul(result_str, NULL, 16); 
                //And I suppose that this is a hex representation of a pointer.
                if(whence == 0){ //SEEK_SET
			unsigned long long ull_offset_hi = (unsigned long long)offset_hi;
                        curr_offset[fd] = (ull_offset_hi<<32) | offset_lo;
                }
                else if(whence == 1){ //SEEK_CUR
			unsigned long long ull_offset_hi = (unsigned long long)offset_hi;
                        curr_offset[fd] += (ull_offset_hi<<32) | offset_lo;
                }
                else if(whence == 2){ //SEEK_END
			unsigned long long ull_offset_hi = (unsigned long long)offset_hi;
                        curr_offset[fd] = ((ull_offset_hi<<32) | offset_lo)+filesize[fd]; 
			//We use filesize because that's the closest to end we know of at the moment.
                }
                else{ print_error("Wrong whence value in _llseek\n"); }
                if(curr_offset[fd] > filesize[fd]){
                        filesize[fd] = curr_offset[fd];
                        #ifdef DEBUG
                        printf("%s filesize updated to %llu\n", paths[fd], filesize[fd]);
                        #endif
                }
        }
        else if((strcmp(callname, "lseek64")==0) && (atoi(retval) >= 0) ){
                #ifdef DEBUG
                printf("lseek64 case\n");
                #endif
                int fd = atoi(strtok(NULL, ","));
                unsigned long long offset = atoi(strtok(NULL, ","));
                int whence = atoi(strtok(NULL, ")"));
                if(whence == 0){ //SEEK_SET
                        curr_offset[fd] = offset;
                }
                else if(whence == 1){ //SEEK_CUR
                        curr_offset[fd] += offset;
                }
                else if(whence == 2){ //SEEK_END
                        curr_offset[fd] = offset+filesize[fd]; 
			//We use filesize because that's the closest to end we know of at the moment.
		}
                else{ print_error("Wrong whence value in _llseek\n"); }
                if(curr_offset[fd] > filesize[fd]){
                        filesize[fd] = curr_offset[fd];
                }
        }
        else if((strcmp(callname, "lseek")==0) && (atoi(retval) >= 0) ){
                #ifdef DEBUG
                printf("lseek case\n");
                #endif
                int fd = atoi(strtok(NULL, ","));
                off_t offset = atoi(strtok(NULL, ","));
                int whence = atoi(strtok(NULL, ")"));
                if(whence == 0){ //SEEK_SET
                        curr_offset[fd] = offset;
                }
                else if(whence == 1){ //SEEK_CUR
                        curr_offset[fd] += offset;
                }
                else if(whence == 2){ //SEEK_END
                        curr_offset[fd] = offset+filesize[fd]; 
			//We use filesize because that's the closest to end we know of at the moment.
                }
                else{ print_error("Wrong whence value in _llseek\n"); }
                if(curr_offset[fd] > filesize[fd]){
                        filesize[fd] = curr_offset[fd];
                }
        }
return 0;
}

int reconstruct_fs(FILE *trace){
        //We construct the file on close because the next open may reuse the fd,
        //and we may reopen the first file again after closing the second file, again getting the same fd.

        char *timestamp;
        char *call;
        char *retval;
        char line[LINELEN];

        fgets(line, LINELEN, trace);
        while(!feof(trace)){
                timestamp = strtok(line, "  ");
                call = strtok(NULL, ")");
                retval = strtok(NULL, " "); //to kill the = sign
                retval = strtok(NULL, " ");
                while(isspace(*call)){ call++; } //Trim the leading whitespace. Not sure why it's there in the first place though.
                gather_fsstat(call, retval);
                fgets(line, LINELEN, trace);
        }
        rewind(trace);
	//Clean up our fsstats because sometimes files aren't closed at the end of the trace.
	//We'll go through filesize array for any sizes greater than 0.
	int i;
	for(i=0; i<MAXFILES; i++){
		if(filesize[i] > 0){ //If an entry has a filesize greater than zero, check if we have a pathname for it.
			if(paths[i] != NULL){	
				char path_copy[strlen(paths[i])];
				strcpy(path_copy, paths[i]);
				if(chdir(prefix) != 0 ) {print_error("Failed to change to target\n"); }
				char *first = strtok(paths[i], "/");
				char initial[strlen(prefix)+strlen(first)];
				strcpy(initial, prefix);
				strcat(initial, first);
				char *path_curr = initial;
				char *path_next = strtok(NULL, "/");
				#ifdef CREATION_DEBUG
				printf("Path to create: %s\n", path_copy);
				#endif
				while(path_next != NULL){ //While next is still valid, path_curr holds a directory name.
					#ifdef CREATION_DEBUG
					printf("curr: %s, next: %s\n", path_curr, path_next);
					#endif
					//Make a dir called path_curr
					if( (mkdir(path_curr, S_IRWXU|S_IRWXO|S_IRWXG) != 0 ) && 
					    (errno != EEXIST) ){ print_error("mkdir failed\n"); }
					#ifdef CREATION_DEBUG
					printf("Made dir\n");
					#endif
					if(chdir(path_curr) != 0 ) {print_error("Failed to change to new directory\n"); }
					path_curr = path_next;
					path_next = strtok(NULL, "/");
				} //On exit of the loop, path_curr holds the last valid path part, so make a file of that name.
		                FILE *new_file = fopen(path_curr, "r");
		                if(new_file == NULL){
		                        #ifdef CREATION_DEBUG
                		        printf("File %s does not exist\n", path_curr); sleep(3);
                		        printf("Filesize will be %llu\n", filesize[i]);
                        		#endif
		                        new_file = fopen(path_curr, "w");
		                        int j;
		                        for(j=0; j< filesize[i]; j++){ fwrite("0", 1, 1, new_file); }
		                        fclose(new_file);
		
		                        #ifdef CREATION_DEBUG
		                        printf("Created %s, with filesize %llu\n", path_copy, filesize[i]);
		                        #endif
		                }
		                else{ //The file already exists so compare current filesize with newly computed filesize.
		                        struct stat statbuf;
		                        #ifdef CREATION_DEBUG
		                        printf("File %s already exists, checking filesize\n", path_curr);
		                        #endif
		                        if(fstat(fileno(new_file), &statbuf) == 0){
		                                fclose(new_file);
		                                #ifdef CREATION_DEBUG
		                                printf("curr size: %llu, computed: %llu\n", (unsigned long long)statbuf.st_size, filesize[i]);
		                                #endif
		                                if((unsigned long long)statbuf.st_size < filesize[i]){ 
							//Computed size is greater than current size so append the file.
		                                        unsigned long long diff = filesize[i] - (unsigned long long)statbuf.st_size;
		                                        #ifdef CREATION_DEBUG
		                                        printf("appending %llu bytes\n", diff);
		                                        #endif
		                                        new_file = fopen(path_curr, "a");
		                                        int k;
		                                        for(k=0; k< diff; k++){ fwrite("0", 1, 1, new_file); }
		                                        fclose(new_file);
		                                }
		                                #ifdef CREATION_DEBUG
		                                else{ printf("%s is already large enough\n", path_curr); }
		                                #endif
		                        }
		                        else{ print_error("fstat failed\n"); }
                		}
				//And clear the stats associated with that fd.
				free(paths[i]);
				paths[i] = NULL;
				curr_offset[i] = 0;
				filesize[i] = 0;
				if(chdir(prefix) != 0 ) {print_error("Failed to change to target\n"); }
			}
			else{ printf("We have a nonzero filesize but no name for it...\n"); }
		}
	}
return 0;
}


int main(int argc, char *argv[]){
	//Parse command line arguments.
	int flag_count = 0;
	int mpi = 0;
	int nofs = 0;
	int drop = 0;
	int a;
	for(a=1; a<argc-1; a++){
		if(!strcmp(argv[a], "-mpi")){
			flag_count++;
			mpi++;
		}
		else if(!strcmp(argv[a], "-nofs")){
			flag_count++;
			nofs++;
		}
		else if(!strcmp(argv[a], "-drop")){
			flag_count++;
			drop++;
		}

	}
	int num_traces = argc-2-flag_count;
	
	char temp_path[1024];
	memset(temp_path, '\0', 1024);
	strncpy(temp_path, argv[1+flag_count], strlen(argv[1+flag_count]));
	if(temp_path[strlen(argv[1+flag_count])-1] != '/'){
		temp_path[strlen(argv[1+flag_count])] = '/';
	}
	prefix = temp_path;

	#ifdef MAIN_DEBUG
	printf("target %s, flag count %d\n", prefix, flag_count);
	#endif

	

	char base_dir[1024];
	getcwd(base_dir, 1024);

	if(drop){ //We try it up here first, so we can exit before a lengthy FS
		  //reconstruction, if drop was flagged without root access.
		//If enabled, we drop the file cache before running tests.
	        sync();
        	FILE *proc = fopen("/proc/sys/vm/drop_caches", "w");
	        if(proc){
        	        if( fwrite("3", 1, 1, proc) != 1){ print_error("Failed to drop caches\n"); exit(-1);}
                	fclose(proc);
	        }
        	else{ print_error("Failed to open proc\n"); exit(-1); }
	}

	if(nofs){} //do nothing
	else{
		//For each input trace, reconstruct the FS.
	        int h;
	        for(h=2+flag_count; h<argc; h++){
			#ifdef RECON_DEBUG
			printf("Starting reconstruct %d of %d\n", h-1-flag_count, num_traces);
			printf("%s\n", argv[h]);
			#endif
	       		FILE *trace;
	                trace = fopen(argv[h], "r");
	                if( trace==NULL){
				printf("Trace %s ", argv[h]);
	                        print_error("failed to open.");
	                }
	                else{	
				#ifdef RECON_DEBUG
				printf("Reconstructing %s\n", argv[h]);
				#endif
	 	       		reconstruct_fs(trace);
				if(fclose(trace)==-1){print_error("Failed to close trace\n");}
				#ifdef RECON_DEBUG
				printf("Reconstructed %s\n", argv[h]);
				#endif
				if(chdir(base_dir) != 0 ) {print_error("Failed to change to back to executing directory\n"); }
                	}
        	}
		int k;
		for(k=0; k<MAXFILES; k++){ fd_arr[k] = -1; }
	        printf("-----------------------------------Finished reconstructing FS\n");
	}

	if(drop){
		//If enabled, we drop the file cache before running tests.
	        sync();
        	FILE *proc = fopen("/proc/sys/vm/drop_caches", "w");
	        if(proc){
        	        if( fwrite("3", 1, 1, proc) != 1){ print_error("Failed to drop caches\n"); exit(-1);}
                	fclose(proc);
	        }
        	else{ print_error("Failed to open proc\n"); exit(-1); }
	}


	if(mpi){
		char mpi_path[] = "/usr/local/bin/mpiexec";
		struct stat mpi_stat;
		if(stat(mpi_path, &mpi_stat) == -1){
			printf("Expected to find mpiexec at %s\n", mpi_path);
			print_error("Did not find mpiexec\n"); exit(-1);
		}
		pid_t pid;
		int status;
		pid = fork();
		if(pid == 0){ //child
			int child_num_flags = 1;
			char *child_argv[6+child_num_flags+num_traces];
			//4 for the mpiexec call: mpiexec -n num_procs app
			//1 for the target
			//1 for the NULL
			//whatever flags we need to pass down from here
			//all of the traces we received
			child_argv[0]=mpi_path;
			child_argv[1]="-n";
			char buf[2];
			sprintf(buf, "%d", num_traces);
			child_argv[2]=buf; //"32";
	
			
			child_argv[3]="./traceplay_mpi";
			child_argv[4]="-nofs";
			//if we're doing an MPI-synchronized, multi-process run,
			//the FS, has already been made here, and we need to tell
			//the trace players that we're using MPI.

			//if(drop){ child_argv[6]="-drop"; }
			//Do we need to pass the child the drop flag?
			//If we receive it here, we already dropped.
			//If we're calling from tracecontrol, all the FSes are already
			//made. I don't think this is necessary.

			child_argv[4+child_num_flags]=prefix; //target device
			int j;
			for(j=0; j<num_traces; j++){
				printf("trace file %d: %s\n", j+1, argv[j+2+flag_count]);
				child_argv[j+5+child_num_flags]=argv[j+2+flag_count];
			}
			child_argv[5+child_num_flags+num_traces]=NULL;
			char *cmd = mpi_path;
			if(execvp(cmd, child_argv) == -1){ 
				print_error("Failed to execute mpiexec\n");
				printf("cmd: %s\n", cmd);
				printf("child argv:\n");
				int b;
				for(b=0; b<argc+4; b++){
					printf("%s ", child_argv[b]);
				}
				exit(-1);
			}
		}
		while(waitpid(pid, &status, 0) == -1){}
		if(!WIFEXITED(status) || WEXITSTATUS(status) != 0){
			perror("child terminated with error\n"); 
		}
	}
	else{ //non-MPI version: fork a traceplay for each trace
		#ifdef MAIN_DEBUG
		printf("forking %d children\n", num_traces);
		#endif
		pid_t pid[num_traces];
		int status;
		int i;
	        for(i=0; i<num_traces; i++){
			pid[i] = fork();
			if(pid[i] == 0){ //child
				char cmd[] = "./traceplay";
				char *child_argv[5];
				child_argv[0]="traceplay";
				child_argv[1]="-nofs"; //FS should always be made from the highest level of control
				child_argv[2]=prefix; //target device
				child_argv[3]=argv[i+2+flag_count]; //trace file
				child_argv[4]=NULL;
				if(execvp(cmd, child_argv) == -1){ 
					int z;
					for(z=0; z<5+drop; z++){
						printf("arg %d: %s\n", z, child_argv[z]);
					}
					print_error("Failed to execute\n");
					exit(0);
				}
			}
		}
		int j;
		for(j=0; j<num_traces; j++){
			while(waitpid(pid[j], &status, 0) == -1){}
			if(!WIFEXITED(status) || WEXITSTATUS(status) != 0){
				perror("child terminated with error\n"); 
			}
		}
	}

return 0;
}
