#include<stdlib.h>
#include<stdio.h>
#include<unistd.h>
#include<errno.h>
#include<string.h>
#include<time.h>
#include<fcntl.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<ctype.h>
#include<sys/time.h>

#ifdef MPI_RUN
#include<mpi.h>
#endif

#define TIMELEN 15
#define CALLLEN 80
#define RETVALLEN 10
#define LINELEN 200
#define MAXFILES 50

int fd_arr[MAXFILES] = {-1}; //It seems bad to hard code this, but I don't feel like doing a dynamically sized array right now.
char *paths[MAXFILES]; //Path/filenames
unsigned long long curr_offset[MAXFILES] = {0};
unsigned long long filesize[MAXFILES] = {0};
char *prefix; //Target mount point

//What stats should we collect?
double read_IO = 0.0; // bytes read
double write_IO = 0.0; //bytes written
struct timeval global_start[1], global_stop[1], read_start[1], read_stop[1], write_start[1], write_stop[1];
double global_elapsed;
int num_read = 0;
int num_write = 0;

int rank, size;

#ifdef LINES
int linecount = 0;
int progress = 5000;
int increment = 5000;
#endif

void print_error(char *msg){
	printf("Error: %s - %s\n", msg, strerror(errno));
	exit(-1);
}

double diff_elapsed(struct timeval start[1], struct timeval stop[1]){
        if( start->tv_usec > stop->tv_usec){
                return (double)(stop->tv_sec - start->tv_sec - 1)+((double)(1000000 + stop->tv_usec - start->tv_usec)/1000000.0);
        }
        else{
                return (double)(stop->tv_sec - start->tv_sec)+((double)(stop->tv_usec - start->tv_usec)/1000000.0);
        }
}

//Count number of arguments in a trace call.
int parm_count(char call[]){
	int count = 1;
	char *mark = strpbrk(call, ",");
	while (mark != NULL){
		count++;
		mark = strpbrk(mark+1, ",");
	}
	return count;
}

int gather_fsstat(char call[], char retval[]){
	char callcopy[strlen(call)];
	strcpy(callcopy, call);
	char *callname=strtok(call, "(");
	#ifdef DEBUG
	printf("call name: %s\n", callname);
	printf("ret val: %s\n", retval);
	#endif
	if((strcmp(callname, "open")==0) && (atoi(retval) > 0)){
		#ifdef DEBUG
		printf("open case\n");
		#endif
		int old_fd = atoi(retval);
		if(old_fd >=50){ print_error("Tried to open more than MAXFILES files. Increase the limit.\n"); }
		char *path_str = strtok(NULL, ",");
		int pathlen = strlen(path_str)-1; //Shave off the quotes but need to add a terminator.
		char path[pathlen];
		memcpy(path, path_str+1, pathlen-1);
		memcpy(&path[pathlen-1], "\0", 1); 
		paths[old_fd] = (char *)malloc(pathlen);
		memcpy(paths[old_fd], path, pathlen);
		#ifdef DEBUG
		
		#endif
	}
	else if( (strcmp(callname, "close")==0) && (atoi(retval)==0) ){
		//We do file creation and checking on close. Rather than a complete static analyis,
		//piecemeal file creation works best in case of fd reuse and fd oddities in the traces.
		int fd = atoi(strtok(NULL, ")"));
		if(paths[fd] != NULL){ //The LANL traces call close multiple, consecutive times on an fd successfully...
			//Create file at paths[fd].
			#ifdef CREATION_DEBUG
			printf("Path to create: %s\n", paths[fd]);
			#endif
			char path_copy[strlen(paths[fd])];
			strcpy(path_copy, paths[fd]);
			if(chdir(prefix) != 0 ) {print_error("Failed to change to target\n"); }
			char *first = strtok(paths[fd], "/");
			char initial[strlen(prefix)+strlen(first)];
			memset(initial, '\0', strlen(prefix)+strlen(first));
			strncpy(initial, prefix, strlen(prefix));
			strncat(initial, first, strlen(first));
			char *path_curr = initial;
			char *path_next = strtok(NULL, "/");
			while(path_next != NULL){ //While next is still valid, path_curr holds a directory name.
				#ifdef CREATION_DEBUG
				printf("curr: %s, next: %s\n", path_curr, path_next);
				#endif
				//Make a dir called path_curr
				if( (mkdir(path_curr, S_IRWXU|S_IRWXO|S_IRWXG) != 0 ) && (errno != EEXIST) ){ print_error("mkdir failed\n"); }
				#ifdef CREATION_DEBUG
				printf("Made dir\n");
				#endif
				if(chdir(path_curr) != 0 ) {print_error("Failed to change to new directory\n"); }
				path_curr = path_next;
				path_next = strtok(NULL, "/");
			} //On exit of the loop, path_curr holds the last valid path part, so make a file of that name.
			FILE *new_file = fopen(path_curr, "r");
			if(new_file == NULL){ 
				#ifdef CREATION_DEBUG
				printf("File does not exist\n"); 
				printf("Filesize will be %llu\n", filesize[fd]);
				sleep(3);
				#endif
				new_file = fopen(path_curr, "w");
				unsigned long long i;
				for(i=0; i< filesize[fd]; i++){ fwrite("0", 1, 1, new_file); }
				fclose(new_file);

				#ifdef CREATION_DEBUG
				printf("Created %s, with filesize %llu\n", path_copy, filesize[fd]);
				#endif
			}
			else{ //The file already exists so compare current filesize with newly computed filesize.
				struct stat statbuf;
				#ifdef CREATION_DEBUG
				printf("file already exists, checking filesize\n");
				#endif
				if(fstat(fileno(new_file), &statbuf) == 0){
				fclose(new_file);
				#ifdef CREATION_DEBUG
				printf("curr size: %llu, computed: %llu\n", (unsigned long long)statbuf.st_size, filesize[fd]);
				#endif
				if((unsigned long long)statbuf.st_size < filesize[fd]){ 
					//Computed size is greater than current size so append the file.
					unsigned long long diff = filesize[fd] - (unsigned long long)statbuf.st_size;
					#ifdef CREATION_DEBUG
					printf("appending %llu bytes\n", diff);
					#endif
					new_file = fopen(path_curr, "a");
					unsigned long long i;
					for(i=0; i< diff; i++){ 
						
						int ret = fwrite("0", 1, 1, new_file); 
						if( (ret < 1) && (ferror(new_file)) ){
							printf("didn't append\n");
						}
					}
					fclose(new_file);
				}
				#ifdef CREATION_DEBUG
				else{ printf("%s is already large enough\n", path_curr); }
				#endif
				}
				else{ print_error("fstat failed\n"); }
			}
			//And clear the stats associated with that fd.
			free(paths[fd]);
			paths[fd] = NULL;
			curr_offset[fd] = 0;
			filesize[fd] = 0;
			if(chdir(prefix) != 0 ) {print_error("Failed to change to target\n"); }
		}
	}
	else if(( (strcmp(callname, "read")==0) || (strcmp(callname, "readv")==0) || 
		  (strcmp(callname, "write")==0) || (strcmp(callname, "writev")==0)) && 
		  (atoi(retval) >= 0) ){
		#ifdef DEBUG
		printf("read/readv/write/writev case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		if(paths[fd] !=NULL){//As long as we're not writing to one the stds
			curr_offset[fd] += atoi(retval);
			if(curr_offset[fd] > filesize[fd]){
				filesize[fd] = curr_offset[fd];
				#ifdef DEBUG
				printf("%s filesize updated to %llu\n", paths[fd], filesize[fd]);
				#endif
			}
		}

	}
	else if( (strcmp(callname, "_llseek") == 0) && (atoi(retval) >= 0) ){ 
		//Seeks normally return the resulting offset but return zeros in the traces.
		#ifdef DEBUG
		printf("_llseek case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		char *offset_hi_str = strtok(NULL, ",");
		char *offset_lo_str = strtok(NULL, ",");
		strtok(NULL, ","); //Was result_str
		int whence = atoi(strtok(NULL, ")"));
		unsigned long offset_hi = strtoul(offset_hi_str, NULL, 10); //I'm guessing this argument is a decimal.
		unsigned long offset_lo  = strtoul(offset_lo_str, NULL, 16); //But it's strange that this one is given in hex.
		//unsigned long result = strtoul(result_str, NULL, 16); //And I suppose that this is a hex representation of a pointer.
		if(whence == 0){ //SEEK_SET
			unsigned long long ull_offset_hi = (unsigned long long)offset_hi;
			curr_offset[fd] = (ull_offset_hi<<32) | offset_lo;
		}
		else if(whence == 1){ //SEEK_CUR
			unsigned long long ull_offset_hi = (unsigned long long)offset_hi;
			curr_offset[fd] += (ull_offset_hi<<32) | offset_lo;
		}
		else if(whence == 2){ //SEEK_END
			unsigned long long ull_offset_hi = (unsigned long long)offset_hi;
			curr_offset[fd] = ((ull_offset_hi<<32) | offset_lo)+filesize[fd]; 
			//We use filesize because that's the closest to end we know of at the moment.
		}
		else{ print_error("Wrong whence value in _llseek\n"); }
		if(curr_offset[fd] > filesize[fd]){
			filesize[fd] = curr_offset[fd];
			#ifdef DEBUG
			printf("%s filesize updated to %llu\n", paths[fd], filesize[fd]);
			#endif

		}
	}
	else if((strcmp(callname, "lseek64")==0) && (atoi(retval) >= 0) ){
		#ifdef DEBUG
		printf("lseek64 case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		unsigned long long offset = atoi(strtok(NULL, ","));
		int whence = atoi(strtok(NULL, ")"));
		if(whence == 0){ //SEEK_SET
			curr_offset[fd] = offset;
		}
		else if(whence == 1){ //SEEK_CUR
			curr_offset[fd] += offset;
		}
		else if(whence == 2){ //SEEK_END
			curr_offset[fd] = offset+filesize[fd]; 
			//We use filesize because that's the closest to end we know of at the moment.
		}
		else{ print_error("Wrong whence value in _llseek\n"); }
		if(curr_offset[fd] > filesize[fd]){
			filesize[fd] = curr_offset[fd];
		}
	}
	else if((strcmp(callname, "lseek")==0) && (atoi(retval) >= 0) ){
		#ifdef DEBUG
		printf("lseek case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		off_t offset = atoi(strtok(NULL, ","));
		int whence = atoi(strtok(NULL, ")"));
		if(whence == 0){ //SEEK_SET
			curr_offset[fd] = offset;
		}
		else if(whence == 1){ //SEEK_CUR
			curr_offset[fd] += offset;
		}
		else if(whence == 2){ //SEEK_END
			curr_offset[fd] = offset+filesize[fd]; 
			//We use filesize because that's the closest to end we know of at the moment.
		}
		else{ print_error("Wrong whence value in _llseek\n"); }
		if(curr_offset[fd] > filesize[fd]){
			filesize[fd] = curr_offset[fd];
		}
	}
	return 0;
}

int reconstruct_fs(FILE *trace){
	char *timestamp;
	char *call;
	char *retval;
	char line[LINELEN];
	
	
	fgets(line, LINELEN, trace);
	while(!feof(trace)){
		timestamp = strtok(line, "  ");
		call = strtok(NULL, ")");
		retval = strtok(NULL, " "); //to kill the = sign
		retval = strtok(NULL, " ");
		while(isspace(*call)){ call++; } //Trim the leading whitespace. Not sure why it's there in the first place though.

		gather_fsstat(call, retval);		

		fgets(line, LINELEN, trace);
	}
	rewind(trace);
	//Clean up our fsstats because sometimes files aren't closed at the end of the trace.
	//We'll go through filesize array for any sizes greater than 0.
	int i;
	for(i=0; i<MAXFILES; i++){
		if(filesize[i] > 0){ //If an entry has a filesize greater than zero, check if we have a pathname for it.
			if(paths[i] != NULL){	
				char path_copy[strlen(paths[i])];
				strcpy(path_copy, paths[i]);
				if(chdir(prefix) != 0 ) {print_error("Failed to change to target\n"); }
				char *first = strtok(paths[i], "/");
				char initial[strlen(prefix)+strlen(first)];
				memset(initial, '\0', strlen(prefix)+strlen(first));
				strncpy(initial, prefix, strlen(prefix));
				strncat(initial, first, strlen(first));
				char *path_curr = initial;
				char *path_next = strtok(NULL, "/");
				#ifdef CREATION_DEBUG
				printf("Path to create: %s\n", path_copy);
				#endif
				while(path_next != NULL){ //While next is still valid, path_curr holds a directory name.
					#ifdef CREATION_DEBUG
					printf("curr: %s, next: %s\n", path_curr, path_next);
					#endif
					//Make a dir called path_curr
					if( (mkdir(path_curr, S_IRWXU|S_IRWXO|S_IRWXG) != 0 ) && 
					    (errno != EEXIST) ){ print_error("mkdir failed\n");} 
					#ifdef CREATION_DEBUG
					printf("Made dir\n");
					#endif
					if(chdir(path_curr) != 0 ) {print_error("Failed to change to new directory\n"); }
					path_curr = path_next;
					path_next = strtok(NULL, "/");
				} //On exit of the loop, path_curr holds the last valid path part, so make a file of that name.
				FILE *new_file = fopen(path_curr, "r");
				if(new_file == NULL){ 
					#ifdef CREATION_DEBUG
					printf("File does not exist\n"); 
					printf("Filesize will be %llu\n", filesize[i]);
					sleep(3);
					#endif
					new_file = fopen(path_curr, "w");
					unsigned long long j;
					for(j=0; j< filesize[i]; i++){ fwrite("0", 1, 1, new_file); }
					fclose(new_file);
	
					#ifdef CREATION_DEBUG
					printf("Created %s, with filesize %llu\n", path_copy, filesize[i]);
					#endif
				}
				else{//The file already exists so compare current filesize with newly computed filesize.
					struct stat statbuf;
					#ifdef CREATION_DEBUG
					printf("file already exists, checking filesize\n");
					#endif
					if(fstat(fileno(new_file), &statbuf) == 0){
						fclose(new_file);
						#ifdef CREATION_DEBUG
						printf("curr size: %llu, computed: %llu\n", (unsigned long long)statbuf.st_size, filesize[i]);
						#endif
						if((unsigned long long)statbuf.st_size < filesize[i]){ 
							//Computed size is greater than current size so append the file.
							unsigned long long diff = filesize[i] - (unsigned long long)statbuf.st_size;
							#ifdef CREATION_DEBUG
							printf("appending %llu bytes\n", diff);
							#endif
							new_file = fopen(path_curr, "a");
							unsigned long long k;
							for(k=0; k< diff; k++){ fwrite("0", 1, 1, new_file); }
							fclose(new_file);
						}
						#ifdef CREATION_DEBUG
						else{ printf("%s is already large enough\n", path_curr); }
						#endif
					}
					else{ print_error("fstat failed\n"); }
				}
				//And clear the stats associated with that fd.
				free(paths[i]);
				paths[i] = NULL;
				curr_offset[i] = 0;
				filesize[i] = 0;
				if(chdir(prefix) != 0 ) {print_error("Failed to change to target\n"); }
			}
			else{ printf("We have a nonzero filesize but no name for it...\n"); }
		}
	}
	return 0;
}

char *make_call(char call[], char retval[]){
	char callcopy[strlen(call)];
	strcpy(callcopy, call);
	char *callname=strtok(call, "(");
	#ifdef DEBUG
	printf("call name: %s\n", callname);
	printf("ret val: %s\n", retval);
	#endif
	if((strcmp(callname, "open")==0) && (atoi(retval) > 0)){
		#ifdef DEBUG
		printf("open case\n");
		#endif
		int fd;
		int old_fd = atoi(retval);
		if(old_fd >=50){ print_error("Tried to open more than MAXFILES files. Increase the limit.\n"); }
		char *path_str = strtok(NULL, ",");
		int pathlen = strlen(path_str)-1; //Shave off the quotes but need to add a terminator.
		char path[pathlen+strlen(prefix)];
		strcpy(path, prefix);
		if(path_str[1]=='/'){
			memcpy(path+strlen(prefix)-1, path_str+1, pathlen-1);
			memcpy(&path[pathlen-2+strlen(prefix)], "\0", 1); 
		}
		else{
			memcpy(path+strlen(prefix), path_str+1, pathlen-1);
			memcpy(&path[pathlen-1+strlen(prefix)], "\0", 1); 
		}
		char *flags_str, *mode_str;
		int flags;
		mode_t mode;
		int num_parm = parm_count(callcopy);
		if( (num_parm==2) || (num_parm ==3)){
			if(num_parm == 3){
				flags_str = strtok(NULL, ",");
				mode_str = strtok(NULL, ")");
				flags = atoi(flags_str);
				mode = (mode_t)atoi(mode_str);
				fd = open(path, flags, mode);
			}
			else if(num_parm == 2){
				flags_str = strtok(NULL, ")");
				flags = atoi(flags_str);
				fd = open(path, flags);
			}
			if(fd == -1){ 
				printf("%s\n", path);
				print_error("File failed to open\n"); 
			}
			#ifdef DEBUG
			else if((fd==0) ||(fd == 1)){ printf("%s mapped to stdout\n", path); }
			#endif
			else{ //Record the fd info.
				fd_arr[old_fd] = fd;
				#ifdef DEBUG
				printf("old fd %d mapped to new fd %d\n", old_fd, fd_arr[old_fd]);
				#endif
			}
		}
		else{ print_error("Malformed open call?\n"); }
		#ifdef DEBUG
		printf("path: %s, flags: %d, mode: %d\n", path, flags, mode);
		#endif
	}
	else if((strcmp(callname, "close")==0) && (atoi(retval)==0)){
		int fd = atoi( strtok(NULL, ")"));
		#ifdef DEBUG
		printf("fd to close %d\n", fd);
		printf("fd to actually close %d\n", fd_arr[fd]);
		#endif
		if(fd_arr[fd] != -1){
			if(close(fd_arr[fd]) == 0){
				fd_arr[fd] = -1;
			}
			else{ 
				printf("Process %d:", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
				print_error("Failed to close a file.\n");
			} //LANL trace has multiple files mapped to the same fd? 
		}
	}
	else if((strcmp(callname, "read")==0) && (atoi(retval) >= 0) ){
		#ifdef DEBUG
		printf("read case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		if(fd_arr[fd] != -1){
			strtok(NULL, ","); //skip the buffer
			int count = atoi(retval); 
			//We just use the retval because if the call succeeded, count and retval should be equal.
			char read_buf[count];
			#ifdef DEBUG
			printf("read fd: %d, num bytes: %d\n", fd, count);
			#endif
			if(read(fd_arr[fd], read_buf, count) != count){
				printf("Process %d: ", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
				print_error("Read call failed\n");
			}
			else{
				num_read++;
				read_IO += count;
			}
			#ifdef DEBUG
			//printf("buf read: %s\n", buf);
			#endif
		}
	}
	else if((strcmp(callname, "readv")==0) && (atoi(retval) >= 0) ){ 
		//Just using read to read the same number of bytes
		#ifdef DEBUG
		printf("readv case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		int count = atoi(retval);
		if(fd_arr[fd] != -1){
			char buf[count]; //Should this have higher scope to preserve it beyond the read? 
			//Maybe not since we don't care about memory operations.
			if( read(fd_arr[fd], buf, count) != count) {
				printf("Process %d:", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
				print_error("Read (replacing readv) failed\n");
			}
			else{
				num_read++;
				read_IO += count;
			}
		}
	}
	else if((strcmp(callname, "write")==0) && (atoi(retval) >= 0) ){
		#ifdef DEBUG
		printf("write case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		strtok(NULL, ","); //Skip the buffer argument.
		int count = atoi(strtok(NULL, ")"));
		if(fd_arr[fd] != -1){
			//Need to init a buffer here and fill it for the write?
			char write_buf[count];
			memset(write_buf, '1', count);
			#ifdef DEBUG
			printf("write fd: %d (actual: %d), num bytes: %d\n", fd, fd_arr[fd], count);
			#endif
			if(write(fd_arr[fd], write_buf, count) != count){
				printf("Process %d:", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
				print_error("Write call failed\n");
			}
			else{
				num_write++;
				write_IO += count;
			}
			#ifdef DEBUG
			#endif
		}
	}
	else if((strcmp(callname, "writev")==0) && (atoi(retval) >= 0) ){ 
		//Just using write to write the same number of bytes.
		#ifdef DEBUG
		printf("writev case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		int count = atoi(retval);
		if(fd_arr[fd] != -1){
			//Need to init a buffer here and fill it for the write?
			char write_buf[count];
			memset(write_buf, '1', count);
			#ifdef DEBUG
			printf("write fd: %d (actual: %d), num bytes: %d, \n", fd, fd_arr[fd], count);
			#endif
			if(write(fd_arr[fd], write_buf, count) != count){ 
				printf("Process %d:", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
				print_error("Write call (replacing writev) failed\n");}
			else{
				num_write++;
				write_IO += count;
			}
			#ifdef DEBUG
			#endif
		}
	}
	else if((strcmp(callname, "fcntl")==0) && (atoi(retval) >= 0) ){
		#ifdef DEBUG
		printf("fcntl case\n");
		#endif
		int num_parm = parm_count(callcopy);
		if (num_parm == 2){
			int fd = atoi(strtok(NULL, ","));
			int cmd = atoi(strtok(NULL, ","));
			int ret = fcntl(fd_arr[fd], cmd);
			if(fd_arr[fd] != -1){
				if(ret == -1){
					printf("Process %d:", rank);
					#ifdef LINES 
					printf(" on line %d\n", linecount);
					#endif
					print_error("fcntl(2) failed\n"); 
				}
				else if(1){
		
				} //Do we need to check return values for specific cmds?
			}
		}
		else if (num_parm == 3){ //There are two fcntl(3)'s. Can we tell which one from the arguments?
			int fd = atoi(strtok(NULL, ","));
			int cmd = atoi(strtok(NULL, ","));
			long arg = strtol(strtok(NULL, ")"), NULL, 10);
			if(fd_arr[fd] != -1){
				int ret = fcntl(fd_arr[fd], cmd, arg);
				if(ret == -1){ print_error("fcntl(3) failed\n"); }
				else if(1){

				}
			}
		}
		else{ print_error("Malformed fcntl?\n"); }
	}
	else if((strcmp(callname, "fcntl64")==0) && (atoi(retval) >= 0) ){ //fcntl64 has 5 parameters...
		#ifdef DEBUG
		printf("fcntl64 case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		int cmd = atoi(strtok(NULL, ","));
		int arg = atoi(strtok(NULL, ","));
		syscall(221, fd_arr[fd], cmd, arg);
		//int ret = fcntl(fd_arr[fd], cmd, arg);
		//if(ret == -1){print_error("fcntl(5) failed\n"); }
		//else if(1){ }

	}
	else if( (strcmp(callname, "_llseek") == 0) && (atoi(retval) >= 0) ){ 
		//Seeks normally return the resulting offset but return zeros in the traces.
		int fd = atoi(strtok(NULL, ","));
		if(fd_arr[fd] != -1){
			char *offset_hi_str = strtok(NULL, ",");
			char *offset_lo_str = strtok(NULL, ",");
			strtok(NULL, ","); //Was result_str
			int whence = atoi(strtok(NULL, ")"));
			unsigned long offset_hi = strtoul(offset_hi_str, NULL, 10); 
			//I'm guessing this argument is a decimal.
			unsigned long offset_lo  = strtoul(offset_lo_str, NULL, 16); 
			//But it's strange that this one is given in hex.
			//unsigned long result = strtoul(result_str, NULL, 16); 
			//And I suppose that this is a hex representation of a pointer.
			loff_t *res = NULL;
			syscall(140, fd_arr[fd], offset_hi, offset_lo, res, whence); 
			// /usr/src/kernels/2.6.23.17-88.fc7-i686/include/asm/unistd.h: _llseek
			#ifdef DEBUG
			printf("_llseek parms: %d, %lu, %lu, %lu, %d\n", fd, offset_hi, offset_lo, result, whence);
			#endif 
		}
	}
	else if((strcmp(callname, "lseek64")==0) && (atoi(retval) >= 0) ){
		#ifdef DEBUG
		printf("lseek64 case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		if(fd_arr[fd] != -1){
			unsigned long long offset = atoi(strtok(NULL, ","));
			int whence = atoi(strtok(NULL, ")"));
			#ifdef DEBUG
			printf("lseek64 parms: %d, %llu, %d\n", fd, offset, whence);
			#endif
			if(lseek64(fd_arr[fd], offset, whence) == ((off_t)-1)){ 
				printf("Process %d:", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
				print_error("lseek64 failed\n"); 
			}
		}
	}
	else if((strcmp(callname, "lseek")==0) && (atoi(retval) >= 0) ){
		#ifdef DEBUG
		printf("lseek case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		if(fd_arr[fd] != -1){
			long offset = atoi(strtok(NULL, ","));
			int whence = atoi(strtok(NULL, ")"));
			#ifdef DEBUG
			printf("lseek parms: %d, %ld, %d\n", fd, offset, whence);
			#endif
			if(lseek(fd_arr[fd], offset, whence) == ((off_t)-1)){ 
				printf("Process %d:", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
				print_error("lseek64 failed\n"); 
			}
		}
	}
	else if((strcmp(callname, "ftruncate")==0) && (atoi(retval) == 0) ){
		#ifdef DEBUG
		printf("ftruncate case\n");
		#endif
		int fd = atoi(strtok(NULL, ","));
		if(fd_arr[fd] != -1){
			off_t length = atoi(strtok(NULL, ")"));
			if (ftruncate(fd_arr[fd], length) != 0){ 
				printf("Process %d:", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
				print_error("ftruncate failed\n"); 
			}
		}
	}
	else if((strcmp(callname, "access")==0) && (atoi(retval) == 0) ){
		#ifdef DEBUG
		printf("access case\n");
		#endif
		char *path_str = strtok(NULL, ",");
		int mode = atoi(strtok(NULL, ")"));
		access(path_str, mode); //Is there any point to checking the return on this? 
					//Since if permissions do not match mode, it also returns -1.
	}
	else if((strcmp(callname, "sync")==0) && (atoi(retval) == 0) ){
		sync();
	}
	else if((strcmp(callname, "fdatasync")==0) && (atoi(retval) == 0) ){
		int fd = atoi(strtok(NULL, ")"));
		if(fd_arr[fd] != -1){ fdatasync(fd_arr[fd]); }
	}
	else if((strcmp(callname, "fsync")==0) && (atoi(retval) == 0) ){
		int fd = atoi(strtok(NULL, ")"));
		if(fd_arr[fd] != -1){ fsync(fd_arr[fd]); }
	}
	else if(strcmp(callname, "sleep")==0) {//&& (atoi(retval) == 0) ){
		int time = atoi(strtok(NULL, ")"));
		sleep(time);
	}
#ifdef MPI_RUN
	//Handle MPI calls. We assume all the MPI dealing with synchronization succeed,
	//so we didn't parse the return values during processing.
	//The calls we want to handle have implicit or explict synchronization
	else if( (strcmp(callname, "MPI_Barrier")==0) ){ //&& (atoi(retval) == MPI_SUCCESS) ){
		MPI_Comm comm = (MPI_Comm)atoi(strtok(NULL, ",")); 
		//Specs say Barrier takes one arg, but traces contain 5.
		#ifdef MPI_DEBUG
		printf("Process %d reached a barrier.\n", rank);
		#endif
		if( (comm != 91) || (MPI_Barrier(MPI_COMM_WORLD) != MPI_SUCCESS) ){
			printf("Process %d failed", rank);
			#ifdef LINES 
			printf(" on line %d\n", linecount);
			#endif
			printf("Comm %d\n", (int)comm);
			print_error("Barrier failed\n");
		}
	}
	else if( (strcmp(callname, "MPI_Bcast")==0) ){ //&& (atoi(retval) == MPI_SUCCESS) ){
		//There doesn't seem to be a need for any corresponding receives.
		//Since we don't know the context or uses of the Bcast data,
		//there doesn't seem to any synchronization to enforce.
		//MPI_Bcast(void *buf, int count, MPI_Datatype datatype, int root, MPI_Comm comm)
		strtok(NULL, ",");
		int count = atoi(strtok(NULL, ","));
		strtok(NULL, ","); //we skip datatype since we'll just use a constant datatype MPI_BYTE
		int root = atoi(strtok(NULL, ","));
		MPI_Comm comm = (MPI_Comm)atoi(strtok(NULL, ")"));
		void *buf = malloc(sizeof(MPI_BYTE)*count);
		#ifdef MPI_DEBUG
		printf("Process %d doing a bcast\n", rank);
		#endif
		if( (comm != 91) || (MPI_Bcast(buf, 1, MPI_BYTE, root, MPI_COMM_WORLD) != MPI_SUCCESS) ){
			printf("Process %d failed", rank);
			#ifdef LINES 
			printf(" on line %d\n", linecount);
			#endif
			printf("Comm %d\n", (int)comm);
			print_error("Bcast failed\n");
		}
		
	}
	else if( (strcmp(callname, "MPI_Reduce")==0) ){ // && (atoi(retval) == MPI_SUCCESS) ){
		//Reduce gathers at a specified rank and collapses to a single value
		//MPI_Reduce(void *sendbuf, void *recvbuf, int count, MPI_Datatype datatype, 
		//MPI_Op op, int root, MPI_Comm comm)
		//recvbuf is only significant at root. 
		//Trace contains only 5 args.
		//My guess is that root and comm are not included.
		//This is similar to every process doing a send to root
		//who does something with all the values.
		//But there doesn't seem to be any indication about whether this is 
		//blocking or not. Notes says the current implementation is a simple tree.
		//The alternative is to just parse the args, make up matching data based on 
		//MPI_Datatype and do an actual reduce.

		//Nevermind. We don't bother parsing the args because
		//we'll just make our own dummy msg for reduce
		//strtok(NULL, ","); strtok(NULL, ",");
		//int count = atoi(strtok(NULL, ","));
		//MPI_Datatype datatype = (MPI_Datatype)atoi(strtok(NULL, ","));
		//MPI_Op op = (MPI_Op)atoi(strtok(NULL, ")"));
		//int root = atoi(strtok(NULL, ","));
		//MPI_Comm comm = atoi(strtok(NULL, ")"));
		int foo = rank; //Our message is an int, our rank number.
		int *recvbuf = (int *)malloc(sizeof(int));
		#ifdef MPI_DEBUG
		printf("%d doing a reduce\n", rank);
		#endif
		if(MPI_Reduce(&foo, recvbuf, 1, MPI_INT, MPI_MAX, 0, MPI_COMM_WORLD) != MPI_SUCCESS) {
			printf("Process %d failed\n", rank);
				#ifdef LINES 
				printf(" on line %d\n", linecount);
				#endif
			print_error("Reduce failed\n");
		}
		#ifdef MPI_DEBUG
		printf("Process %d finished reduce\n", rank);
		#endif
	}
	else if( (strcmp(callname, "MPI_Send")==0) ){ //&& (atoi(retval) == MPI_SUCCESS) ){
		//MPI_Send(void *buf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm)
		//Trace only contains 5 args.
		/* MPI_FLOAT_INT      ((MPI_Datatype)17)
		   MPI_DOUBLE_INT     ((MPI_Datatype)18)
		   MPI_LONG_INT       ((MPI_Datatype)19)
   		   MPI_SHORT_INT      ((MPI_Datatype)20)
		   MPI_2INT           ((MPI_Datatype)21)
		   MPI_LONG_DOUBLE_INT ((MPI_Datatype)22) */
		//Actually, we may be ok just picking one datatype and using
		//that to fill the messages. After all, it's not really
		//the content of the message that's important, but the
		//synchronization impled by the existence of a message.
		strtok(NULL, ",");
		strtok(NULL, ","); //Was to grab count arg.
		strtok(NULL, ","); //Was to grab MPI_Datatype.
		int dest = atoi(strtok(NULL, ","));
		int tag = atoi(strtok(NULL, ")"));
		int foo = rank; //Our message is an int, our rank.
		#ifdef MPI_DEBUG
		printf("Process %d doing a send to process %d: %d\n", rank, dest, foo);
		#endif
		if(MPI_Send(&foo, 1, MPI_INT, dest, tag, MPI_COMM_WORLD) != MPI_SUCCESS) {
			printf("Process %d failed\n", rank);
			#ifdef LINES 
			printf(" on line %d\n", linecount);
			#endif
			print_error("Send failed\n");
		}
		#ifdef MPI_DEBUG
		printf("Process %d finished sending to process %d\n", rank, dest);
		#endif
	}
	else if( (strcmp(callname, "MPI_Recv")==0) ){ //&& (atoi(retval) == MPI_SUCCESS) ){ 
		//Looks like only rank root does these
		//MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, 
		//int tag, MPI_Comm comm, MPI_Status *status)
		//Trace only contains 5 args.
		/* MPI_FLOAT_INT      ((MPI_Datatype)17)
		   MPI_DOUBLE_INT     ((MPI_Datatype)18)
		   MPI_LONG_INT       ((MPI_Datatype)19)
   		   MPI_SHORT_INT      ((MPI_Datatype)20)
		   MPI_2INT           ((MPI_Datatype)21)
		   MPI_LONG_DOUBLE_INT ((MPI_Datatype)22) */
		strtok(NULL, ",");
		strtok(NULL, ","); //Was to grab count argument
		strtok(NULL, ","); //Was to grab the MPI_Datatype
		int source = atoi(strtok(NULL, ","));
		int tag = atoi(strtok(NULL, ")"));
		MPI_Status status;
		int foo;
		#ifdef MPI_DEBUG
		int source_copy = source;
		printf("%d waiting to recv from %d\n", rank, source);
		#endif
		if(MPI_Recv(&foo, 1, MPI_INT, source, tag, MPI_COMM_WORLD, &status) != MPI_SUCCESS) {
			printf("Process %d failed\n", rank);
			#ifdef LINES 
			printf(" on line %d\n", linecount);
			#endif
			print_error("Recv failed\n");
		}
		#ifdef MPI_DEBUG
		printf("Process %d got a message from process %d:%d\n", rank, source_copy, foo);
		#endif
	}
	else if((strcmp(callname, "MPI_Isend")==0) ||
		(strcmp(callname, "MPI_Ibsend")==0) ||
		(strcmp(callname, "MPI_Issend")==0) ||
		(strcmp(callname, "MPI_Irsend")==0) ||
		(strcmp(callname, "MPI_Irecv")==0) ||
		(strcmp(callname, "MPI_Wait")==0) ||
		(strcmp(callname, "MPI_Sendrecv")==0) ||	
		(strcmp(callname, "MPI_Sendrecv_replace")==0) ){
		//There's also Scatter/Gather operations and their variants
		//We don't have any of these calls, I don't think,
		//but we'll do this just to check, just in case.
		printf("Process %d: I found an unexpected MPI call.\n", rank);
		#ifdef LINES
		printf("Process %d: It was on line %d\n", rank, linecount);
		#endif
	}
#endif //end MPI_RUN
	else{
		#ifdef DEBUG
		printf("default case\n");
		printf("failed call: %s\n", callcopy);
		#endif
	}
return "0";
}

int main(int argc, char *argv[]){
	
	//Parse command line arguments.
	//Expected syntax is ./app flags target traces
	int flag_count = 0;
	int drop = 0;
	int nofs = 0;
	int a;
	for(a =0; a<argc-1; a++){
		if(!strcmp(argv[a], "-drop")){
			flag_count++;
			drop++;
		}
		else if(!strcmp(argv[a], "-nofs")){
			flag_count++;
			nofs++;
		}
	}
	#ifdef MAIN_DEBUG
	printf("received %d flags\n", flag_count);
	#endif

	char temp_path[1024];
	memset(temp_path, '\0', 1024);
	strncpy(temp_path, argv[1+flag_count], strlen(argv[1+flag_count]));
	if(temp_path[strlen(argv[1+flag_count])-1] != '/'){
		temp_path[strlen(argv[1+flag_count])] = '/';
	}
	prefix = temp_path;

	if(drop){ //We try to drop early to check for root access before
		  //going through a lengthy FS reconstruct.
		  //If enabled, we clean the file cache before each replay.
		FILE *proc = fopen("/proc/sys/vm/drop_caches", "w");
		if(proc){
			sync();
			if( fwrite("3", 1, 1, proc) != 1){ print_error("Failed to drop caches\n"); exit(-1);}
			fclose(proc);
		}
		else{ print_error("Failed to open proc\n"); exit(-1);}
	}
	#ifdef MPI_RUN
		MPI_Init(&argc, &argv);
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		MPI_Comm_size(MPI_COMM_WORLD, &size);
	#else
		rank=getpid(); size=1;
		
	#endif

	#ifdef MAIN_DEBUG

	#endif

	//Open all the trace files passed in. 
	int num_traces = argc-2-flag_count;
	FILE *traces[num_traces];
	int i;
	for(i=0; i<num_traces; i++){
		traces[i] = fopen(argv[i+2+flag_count], "r");
		if(traces[i]==NULL){
			printf("%s: ", argv[i+2+flag_count]);
			print_error("Failed to open trace.\n");
		}
		#ifdef MAIN_DEBUG
		else{ printf("Opened %s\n", argv[i+2+flag_count]); }
		#endif
	}
	#ifdef MAIN_DEBUG
	printf("Opened all traces.\n");
	#endif

	//FS reconstruction.
	if(nofs){} //Do nothing.
	else{
		int j;
		for (j=0; j<num_traces; j++){
			reconstruct_fs(traces[j]);
			printf("---------Finished reconstructing FS for %s\n", argv[j+2+flag_count]);
			sync();
		}	
	}


	int q;
	#ifndef MPI_RUN
	for(q=0; q<num_traces; q++){
	#else
	q=rank;
	#endif
		if(drop){ //If enabled, we clean the file cache before each replay.
			FILE *proc = fopen("/proc/sys/vm/drop_caches", "w");
			if(proc){
				sync();
				if( fwrite("3", 1, 1, proc) != 1){ print_error("Failed to drop caches\n"); exit(-1);}
				fclose(proc);
			}
			else{ print_error("Failed to open proc\n"); exit(-1); }
		}
		int k;
		for(k=0; k<MAXFILES; k++){ fd_arr[k] = -1;} //Clear our fd map.
		#ifdef MPI_RUN
		printf("Process %d of %d: replaying trace: %s\n", rank, size, argv[q+2+flag_count]);
		#else
		printf("Process %d: replaying trace: %s\n", rank, argv[q+2+flag_count]);
		#endif
		char *timestamp;
		char *call;
		char *retval;
		char line[LINELEN];
	
		fgets(line, LINELEN, traces[q]); //rank is 0 unless MPI assigned it for parallel replay
		#ifdef LINES
		linecount++;
		#endif
		gettimeofday(global_start, NULL);
		while(!feof(traces[q])){
			timestamp = strtok(line, "  ");
			call = strtok(NULL, ")");
			retval = strtok(NULL, " "); //to kill the = sign
			retval = strtok(NULL, " ");
			while(isspace(*call)){ call++; } //Trim the leading whitespace. Not sure why it's there in the first place though.
			#ifdef DEBUG
			printf("%s:%s:%s\n", timestamp, call, retval);
			#endif
	
			make_call(call, retval);
			fgets(line, LINELEN, traces[q]);
			/* print out where the process is in the trace replay.
      		         * useful when you want to make sure proccess are making progress
                 	 * in a synchronized replay*/
			#ifdef LINES
			linecount++;
			if(linecount > progress){
				printf("Process %d, line %d\n", rank, linecount);
				progress += increment;
			}
			#endif
		}
		//Should we do a sync here before exiting?
		sync();
		#ifdef MPI_RUN
		MPI_Finalize(); 
		#endif
		gettimeofday(global_stop, NULL);
		global_elapsed = diff_elapsed(global_start, global_stop);
		printf("Process %d finished playing %s\n", rank, argv[q+2+flag_count]);	
		printf("Process %d:%s - Total bytes read: %f\n", rank, argv[q+2+flag_count], read_IO);
		printf("Process %d:%s - Total bytes written: %f\n", rank, argv[q+2+flag_count], write_IO);
		printf("Process %d:%s - Percent read/write calls: %f/%f\n", rank, argv[q+2+flag_count], (float)num_read/(float)(num_read+num_write), (float)num_write/(float)(num_read+num_write));
		printf("Process %d:%s - Global elapsed: %f s\n", rank, argv[q+2+flag_count], global_elapsed);
		printf("--------------------------------------------\n");
	
		#ifdef LINES
		linecount = 0;
		progress = 5000;
		#endif
	#ifndef MPI_RUN
	}
	#endif
	return 0;
}
